import sys
from time import time

from OFS.SimpleItem import SimpleItem
from Globals import InitializeClass, PersistentMapping, DTMLFile
from AccessControl import ClassSecurityInfo
from AccessControl.Permissions import delete_objects
from AccessControl import getSecurityManager
from Acquisition import aq_base, aq_inner, aq_parent

from Products.CMFCore.CMFCorePermissions import ManagePortal, AccessContentsInformation, ModifyPortalContent, ReviewPortalContent, View, ManageProperties
from Products.CMFCore.utils import UniqueObject
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.ActionProviderBase import ActionProviderBase
from Products.CMFCore.ActionInformation import ActionInformation
from Products.CMFCore.Expression import Expression
from Products.CMFCore.DirectoryView import addDirectoryViews
from Products.DCWorkflow.utils import modifyRolesForPermission

import re

from SkinTrap import SkinTrap
from PVSPermissions import CreateVersions
from urllib import urlencode as encode

_marker = []  # Create a new marker object.

def manage_addPVSTool(self, REQUEST=None):
    " "
    # raise err if not in portal root
    if hasattr(self, 'portal_url'):
        obj = PVSTool()
        if not hasattr(self, obj.getId()):
            self._setObject(obj.getId(), obj)
        if REQUEST: return self.manage_main(self, REQUEST)
    else:
        raise 'Wrong place', 'The current folder is not a CMF Site. PVS tool can only be added to the root of a CMF Site.'

class PVSTool (UniqueObject, SimpleItem, ActionProviderBase):

    """ Provide versioning mechanism to CMF
    """
    id = 'portal_content_versioning_system'
    meta_type = 'Portal content Versioning System'

    manage_options = ({ 'label' : 'Overview', 'action' : 'manage_overview' }, ) + SimpleItem.manage_options

    manage_overview = DTMLFile('dtml/explainPVS', globals())

    _actions = (ActionInformation(id='about_to_create_version'
                                , title='Edit'
                                , action=Expression(
                                    text="python: object.portal_content_versioning_system.getMyVersionURL(object, object_url)"
                                        )
                                , condition=Expression(
                                    text='python: object.portal_content_versioning_system.allowCreateVersion(object)')
                                , permissions=(CreateVersions, )
                                , category='object'
                                , visible=1
                                 )
                ,
                ActionInformation(id='abandon_versions'
                                , title='Abandon'
                                , action=Expression(
                                    text='string: ${object_url}/abandon_version_form')
                                , condition=Expression(
                                    text='''python: object.version_cluster_id > 0 and
                                        (member and member.has_permission(%s, object)) and
                                        object.getId() != object.portal_content_versioning_system.getTopVersionId(object)
                                        ''' % `ModifyPortalContent`)
                                , permissions=(CreateVersions, )
                                , category='object'
                                , visible=1
                                 )
                ,
                ActionInformation(id='my_versions'
                                , title='My Versions'
                                , action=Expression(
                                    text='string: ${object_url}/list_my_versions')
                                , condition=Expression(
                                    text="python: object.version_cluster_id > 0 and \
                                        len(object.portal_content_versioning_system.getMyVersionCluster(object))")
                                , permissions=(CreateVersions, ReviewPortalContent)
                                , category='object'
                                , visible=1
                                 )
                ,
                ActionInformation(id='list_versions'
                                , title='Versions'
                                , action=Expression(
                                    text='string: ${object_url}/list_versions')
                                , condition=Expression(
                                    text="python: object.version_cluster_id > 0")
                                , permissions=(CreateVersions, ReviewPortalContent)
                                , category='object'
                                , visible=1
                                 )
                ,
               )

    security = ClassSecurityInfo()

    def manage_afterAdd(self, item, container):
        action_tool = getToolByName(self, 'portal_actions')
        action_tool.addActionProvider(self.id)

        #Join PVS skin folder to the skin selection
        skinstool = getToolByName(self, 'portal_skins')

        # Set up skin trap
        if not hasattr(skinstool, '_skin_trap_'):
            skinstool._skin_trap_ = SkinTrap()

        if 'PVS' not in skinstool.objectIds():
            addDirectoryViews(skinstool, 'skins', globals())

        skins = skinstool.getSkinSelections()
        for skin in skins:
            path = skinstool.getSkinPath(skin)
            # insert before content skin
            if path.find('PVS') < 0:
                path = path.split('custom,')
                path = path[0] + 'custom,PVS,' + path[1]
                skinstool.addSkinSelection(skin, path)

        catalog_tool = getToolByName(self, 'portal_catalog')
        if not catalog_tool._catalog.indexes.has_key('version_cluster_id'):
            catalog_tool.manage_addIndex('version_cluster_id', 'FieldIndex')

        if not catalog_tool._catalog.schema.has_key('getId'):
            catalog_tool.addColumn('getId')

        workflow_tool = getToolByName(self, 'portal_workflow')
        for wf in workflow_tool.objectValues():
            if hasattr(wf, 'addManagedPermission') and CreateVersions not in wf.permissions:
                wf.addManagedPermission(CreateVersions)

    def manage_beforeDelete(self, item, container):
        action_tool = getToolByName(self, 'portal_actions')
        action_tool.deleteActionProvider(self.id)

        skinstool = getToolByName(self, 'portal_skins')
        if hasattr(skinstool, '_skin_trap_'):
            delattr(skinstool, '_skin_trap_')

        if 'PVS' in skinstool.objectIds():
            skinstool.manage_delObjects('PVS')
        skins = skinstool.getSkinSelections()
        for skin in skins:
            path = skinstool.getSkinPath(skin)
            if path.find('PVS') >= 0:
                path = path.replace('PVS,', '')
                skinstool.addSkinSelection(skin, path)
        # not to delete the catalog index of 'version_cluster_id'
        # for versions might still exist in the site


    # ----------- util methods for object verioning -------------
    #
    security.declareProtected(View, 'allowCreateVersion')
    def allowCreateVersion(self, obj):
        user = getSecurityManager().getUser()
        if user:
            if getattr(aq_base(obj), 'isAnObjectManager', 0):
                permission = ManageProperties
            else:
                permission = ModifyPortalContent
            return not user.has_permission(permission, obj)
        return 0

    security.declareProtected(AccessContentsInformation, 'getAllVersionIds')
    def getAllVersionIds(self, obj):
        pcat = getToolByName(self, 'portal_catalog')
        rslt = obj.version_cluster_id > 0 and pcat(version_cluster_id = obj.version_cluster_id) or []
        return [x.getId for x in rslt]

    security.declareProtected(AccessContentsInformation, 'getVersionCluster')
    def getVersionCluster(self, obj):
        ids = self.getAllVersionIds(obj)
        versions = [getattr(obj.aq_parent, id) for id in ids]
        tuples = [(ver.version_number, ver) for ver in versions]
        #tuples.sort()
        return [tuple[1] for tuple in tuples]

    security.declareProtected(AccessContentsInformation, 'getMyVersionCluster')
    def getMyVersionCluster(self, obj):
        tuples = self.getVersionCluster(obj)
        my_tuples = [obj for obj in tuples if str(getSecurityManager().getUser()) in obj.users_with_local_role('Owner')]
        return my_tuples

    security.declareProtected(AccessContentsInformation, 'getMyVersionURL')
    def getMyVersionURL(self, object, object_url):
        my_versions = object.portal_content_versioning_system.getMyVersionCluster(object)
        my_version = [obj.id for obj in my_versions if object.portal_membership.checkPermission('Modify portal content', obj)]
        display_message = my_version and \
              'You can only edit one version at a time.  This is the Version which you were last editing.' or \
              'You do not have permission to modify this content, so a new version will be created.'
        complete_url = (my_version and '/'.join(object_url.split('/')[:-1] + my_version) or object_url) \
          + '/' + object.portal_types.getTypeInfo(object).getActionById('edit') + '?%s' % encode(
          {'portal_status_message': display_message})

        member = getSecurityManager().getUser()
        return complete_url

    def _versionHasChildren(self, id, vids):
        " "
        prefix = id + '.'
        return [id for id in vids if id.startswith(prefix)]

    security.declareProtected(AccessContentsInformation, 'getTopVersionID')
    def getTopVersionId(self, obj):
        version_id = obj.getId()
        if version_id.find('~v~') != -1:
            return version_id[:version_id.find('~v~')]
        else:
            return version_id

    def abandonVersion(self, obj):
        if self.portal_membership.checkPermission('Modify portal content', obj):
            container = aq_parent(aq_inner(obj))
            container._delObject(obj.getId())

    def delVersions(self, obj, ids):
        " "
        for id in ids:
            if id.find('~v~') < 0:
                raise 'Delete Error', '%s is a frontend object, therefore cannot be deleted from version management' % id
        # sort id by version length descendingly. so that the child get deleted first
        length_id = [(len(id.split('.')), id) for id in ids]
        length_id.sort()
        length_id.reverse()
        ids = [x[1] for x in length_id]
        container = aq_parent(aq_inner(obj))
#        vids = self.getAllVersionIds(obj)
        for id in ids:
#            if self._versionHasChildren(id, vids):
#                raise 'Delete Error', '%s cannot be deleted because it has children. Please delete its children first.' % `id`
            container._delObject(id)

    def createVersion(self, orig_obj):
        container = aq_parent(aq_inner(orig_obj))
        if orig_obj.version_cluster_id < 0:
            # this object isn't verionized. shape it up
            orig_obj.version_cluster_id = int(time())
            orig_obj.version_number = '1'

        new_id = getIdForNewVersion(orig_obj)
        new_version = orig_obj._getCopy(container)
        # check if folderish
        if hasattr(aq_base(orig_obj), 'objectIds'):
            # strip off all sub-objects
            for id in new_version.objectIds():
                new_version._delOb(id)
            # reset the _objects index
            new_version._objects = ()

            # disable folder_contents view from all users (including manager)
            # dispite the setting of the workflow
            # wrap new_context with acquisition first
            nv = new_version.__of__(orig_obj)
            modifyRolesForPermission(nv, 'List folder contents', ())

        new_version._setId(new_id)
        new_version.version_number = new_id.split('~v~')[-1]
        container._setOb(new_id, new_version)
        new_context = container._getOb(new_id)

        # plone and azcat formaker will rename the obj if not set
        self.REQUEST.set('id', new_id)

        # set ownership
        user = getSecurityManager().getUser()
        if user is not None:
            name = user.getUserName()
            if name != 'Anonymous User':
                new_context.__ac_local_roles__ = {name : ['Owner']}

        new_context.notifyWorkflowCreated() # set new version into the default state
        orig_obj.reindexObject() # expose the object as versionized to the catalog
        new_context.reindexObject()
        return new_context


def getIdForNewVersion(obj):
    pvs = getToolByName(obj, 'portal_content_versioning_system')
    ids = pvs.getAllVersionIds(obj)
    prefix = obj.getId()
    if prefix.find('~v~') < 0:
        if hasattr(obj, 'version_number'):
            prefix += '~v~' + obj.version_number
        else:
            prefix += '~v~1'

    child_version_ids = [id for id in ids if re.match(prefix + '.\d+\Z', id)]
    id_suffixes = [int(id.split('.')[-1]) for id in child_version_ids]
    if id_suffixes:
        new_suffix = max(id_suffixes) + 1
    else:
        new_suffix = 1
    return '%s.%d' % (prefix, new_suffix)

InitializeClass(PVSTool)

