import Globals, re
from OFS.SimpleItem import SimpleItem
from Products.CMFCore.utils import getToolByName, _checkPermission
from Acquisition import aq_chain, aq_parent, aq_inner, aq_base, Implicit, Explicit, Acquired
from time import time
from AccessControl import getSecurityManager, ClassSecurityInfo

from ZPublisher.mapply import mapply
from ZPublisher.Publish import call_object, missing_name, dont_publish_class

from Products.CMFCore.CMFCorePermissions import AccessContentsInformation, ModifyPortalContent, View, ReviewPortalContent
from Products.CMFCore.WorkflowCore import WorkflowException
from PVSPermissions import CreateVersions

def relayScriptInvocation(new_context, script_name, REQUEST, upper_skin):
    # find the next script along the acquisition path, call it with the new context
    script = upper_skin.aq_acquire(script_name)
    script = script.__of__(new_context)
    return mapply(script, REQUEST.args, REQUEST,
                    call_object,1,
                    missing_name,
                    dont_publish_class,
                    REQUEST, bind=1)

def editActionTrap(self, container, REQUEST, upper_skin):
    "Hijack the edit action"

    # plone lets user to edit content id in edit form
    # disallow that if the content is a hidden version
    if self.getId().find('~v~') > 0:
        REQUEST.set('id', self.getId()) # no id changes

    pvs = getToolByName(self, 'portal_content_versioning_system')
    if pvs is not None and pvs.allowCreateVersion(self):
        new_context = pvs.createVersion(self)
    else:
        new_context = self

    return relayScriptInvocation(new_context, self.script_name, REQUEST, upper_skin)

def contentStatusModify(self, container, REQUEST, upper_skin):
    "Hijack the publish workflow action"

    if self.getId().find('~v~') > 0 \
       and _checkPermission(ReviewPortalContent, self) \
       and REQUEST.workflow_action == 'publish' :

        to_be_published = aq_base(self)
        container._delObject(to_be_published.getId())

        front_end_id = to_be_published.getId().split('~v~')[0]
        to_be_archived = aq_base(container._getOb(front_end_id))
        to_be_archived_id = front_end_id + '~v~' + to_be_archived.version_number

        # switch ids
        to_be_archived._setId(to_be_archived_id)
        to_be_published._setId(front_end_id)

        # if folderish
        if getattr(to_be_published, 'isAnObjectManager', 0):
            for name, subob in to_be_archived.objectItems():
                to_be_published._setOb(name, aq_base(subob))
                to_be_archived._delOb(name)
            to_be_published._objects = to_be_archived._objects
            to_be_archived._objects = ()
            # passing 'self' because self has acquisition wrapper
            # passing [] as roles, so that who has the permission is acquired from above
            modifyRolesForPermission(self, 'List folder contents', [])

        # switch objs
        container._setOb(to_be_archived.getId(), to_be_archived)
        container._setOb(front_end_id, to_be_published)

        the_archived = container._getOb(to_be_archived_id)
        new_context = container._getOb(front_end_id)

        workflow_tool = getToolByName(self, 'portal_workflow')
        # check if the archived item is published
        review_state = workflow_tool.getInfoFor(the_archived, 'review_state')
        if review_state == 'published':
            # try perform 'archive' transition
            try:
                workflow_tool.doActionFor(the_archived, 'archive')
            except WorkflowException:
                # the associated workflow is not versionized
                # set the archived into default state
                the_archived.notifyWorkflowCreated()

        # force to reindex everything on the two objects
        # the workflow methods, such as notifyWorkflowCreated, doActionFor etc.
        # will only reindex workflow related variables
        the_archived.reindexObject()
        new_context.reindexObject()
    else:
        new_context = self

    return relayScriptInvocation(new_context, 'content_status_modify', REQUEST, upper_skin)

class ExternalSkinScript(Explicit, SimpleItem):
    "A wrapper over the mothods above"
    security = ClassSecurityInfo()

    _v_upper_skin = None

    def __init__(self, function):
        self._func = function

    def setUpperSkinData(self, upper_skin):
        self._v_upper_skin = upper_skin

    # at this point, everyone should be able to call the script
    # it's up to the target script to control who can run it
    security.declareProtected(View, '__call__' )
    def __call__(self):
        skinstool = getToolByName(self, 'portal_skins')
        skintrap = skinstool._skin_trap_
        upper_skin = skintrap._v_upper_skin[0]
        REQUEST = self.REQUEST

        context = self.aq_parent.this() # the new context is the cmf obj this script attaches to
        container = aq_parent(aq_inner(context))
        return apply(self._func, (context, container, REQUEST, upper_skin))


class SkinTrap(SimpleItem):
    "A virtual skin layer that will intercepts the skin script acquicition"

    meta_type = 'Skin Trap'
    id = '_skin_trap_'
    _v_upper_skin = None # declare it as class var so that it can be accessed directly, instead of __dict__

    _contentStatusModify = ExternalSkinScript(contentStatusModify)
    _editActionTrap = ExternalSkinScript(editActionTrap)

    def backUpSkin(self, upper_skin):
        if upper_skin:
            self._v_upper_skin = (upper_skin, ) # wrap it in a list so as to keep the acquisition chain

    def __getattr__(self, name):
        " "
        if name == 'content_status_modify':
            return self._contentStatusModify

        upper_skin = self._v_upper_skin[0]
        # only interested in callings to python script which ends with '_edit', not starts with 'validate'
        if hasattr(upper_skin, name) and name.endswith('_edit') and not name.startswith('validate'):
            attr = getattr(upper_skin, name)
            if hasattr(attr, 'meta_type') and attr.meta_type.endswith('Script (Python)'):
                self.script_name = name
                return self._editActionTrap
        raise AttributeError, name

Globals.InitializeClass(SkinTrap)
